import enum
from typing import List

from . import check, util, word_list


class SessionStatus(enum.Enum):
  ACTIVE = 1
  VICTORY = 2
  DEFEAT = 3
  ERROR = 4


class Session():

  def __init__(self, target: str, total_guesses: int,
               words: word_list.WordList):
    try:
      check.validate_word(target, words)
      if total_guesses <= 0:
        raise ValueError("Expected total_guesses to be positive. "
                         f"Got: {total_guesses}")
      self._target = util.clean_text(target)
      self._total_guesses = total_guesses
      self._status = SessionStatus.ACTIVE
      self._guess_history = []
      self._hint_history = []
      self._words = words
    except ValueError as e:
      self._status = SessionStatus.ERROR
      raise e

  @property
  def status(self):
    return self._status

  @property
  def words(self):
    return self._words

  @property
  def target(self):
    return self._target

  @property
  def total_guesses(self):
    return self._total_guesses

  @property
  def guess_count(self):
    return len(self._guess_history)

  @property
  def remaining_guesses(self):
    return self.total_guesses - self.guess_count

  @property
  def guess_history(self):
    return self._guess_history

  @property
  def hint_history(self):
    return self._hint_history

  def guess(self, guess: str) -> List[check.HintCategory]:
    if self.status != SessionStatus.ACTIVE:
      raise ValueError("Cannot make guesses in an inactive session. "
                       f"State: {self.status}")
    if self.remaining_guesses <= 0:
      raise ValueError("No guesses remaining.")
    guess = util.clean_text(guess)
    hint = check.get_hints(guess, self.target, self.words)
    self._guess_history.append(guess)
    self._hint_history.append(hint)
    if check.is_correct(guess, self.target):
      self._status = SessionStatus.VICTORY
    elif self.remaining_guesses == 0:
      self._status = SessionStatus.DEFEAT
    return hint



import enum
import string
from typing import List

from . import util, word_list


class HintCategory(enum.Enum):
  MISS = 1
  CLOSE = 2
  HIT = 3


def get_hints(guess: str, target: str,
              words: word_list.WordList) -> List[HintCategory]:
  """Returns hint results for a guess."""
  guess = util.clean_text(guess)
  target = util.clean_text(target)
  validate_guess(guess, len(target), words)
  hints = [None for _ in guess]
  # Do hits first.
  for i, (guess_c, target_c) in enumerate(zip(guess, target)):
    if guess_c == target_c:
      hints[i] = HintCategory.HIT
  # This should be a list because we may have duplicates.
  remaining_target_chars = [target[i] for i, h in enumerate(hints) if h is None]
  for i, guess_c in enumerate(guess):
    if hints[i] is not None:
      continue
    if guess_c in remaining_target_chars:
      hints[i] = HintCategory.CLOSE
      remaining_target_chars.remove(guess_c)
  # Now, nones are misses.
  for i, h in enumerate(hints):
    if h is None:
      hints[i] = HintCategory.MISS
  return hints


def is_correct(guess: str, target: str) -> bool:
  return util.clean_text(guess) == util.clean_text(target)


_VALID_CHARS = {c for c in string.ascii_uppercase}


def validate_guess(guess: str, target_len: int, words: word_list.WordList):
  """Returns true if the guess is only `target_len` alpha chars."""
  guess = util.clean_text(guess)
  if len(guess) != target_len:
    raise ValueError(f"The guess '{guess}' should be {target_len} "
                     "characters long.")
  validate_word(guess, words)


def validate_word(word: str, words: word_list.WordList):
  assert word is not None
  word = util.clean_text(word)
  if not word:
    raise ValueError("The empty string is not a valid word.")
  if word not in words:
    raise ValueError(f"'{word}' does not appear in the dictionary.")
  if any(c not in _VALID_CHARS for c in word):
    raise ValueError(f"The word'{word}' should only contain letters.")


def validate_char(char: str):
  char = util.clean_text(char)
  if len(char) != 1:
    raise ValueError(f"'{char}' is not a single char.")
  if char not in _VALID_CHARS:
    raise ValueError(f"'{char}' is not a valid letter.")



from typing import List

from rich import align, box, layout, panel, table, text
from textual import app, reactive, widget

from . import check, session, util, word_list


def get_guess_text(letter: str, hint: check.HintCategory) -> text.Text:
  guess_text = text.Text(letter)
  if hint == check.HintCategory.HIT:
    guess_text.stylize("bold green")
  elif hint == check.HintCategory.MISS:
    guess_text.stylize("bright_black")
  elif hint == check.HintCategory.CLOSE:
    guess_text.stylize("bold yellow")
  else:
    raise NotImplementedError(f"Cannot stylize {hint}")
  return guess_text


class GuessTable(widget.Widget):

  def __init__(self, sess: session.Session):
    super().__init__()
    self._session = sess

  @property
  def total_guesses(self) -> int:
    return self._session.total_guesses

  @property
  def target_length(self) -> int:
    return len(self._session.target)

  @property
  def guess_history(self) -> List[str]:
    return self._session.guess_history

  @property
  def hint_history(self) -> List[check.HintCategory]:
    return self._session.hint_history

  def render(self):
    guess_table = table.Table(
        show_header=False, show_footer=False, show_lines=True, box=box.SQUARE)
    for _ in range(self.target_length):
      guess_table.add_column(justify="center", width=1)
    for idx in range(self.total_guesses):
      if idx < len(self.guess_history):
        guess = self.guess_history[idx]
        hint = self.hint_history[idx]
        texts = [get_guess_text(g, h) for g, h in zip(guess, hint)]
        guess_table.add_row(*texts)
      else:
        empty_guess = [" "] * self.target_length
        guess_table.add_row(*empty_guess)
    return align.Align(guess_table, align="center")


class PendingGuessPanel(widget.Widget):
  pending_guess = reactive.Reactive("")

  def __init__(self, target_length: int):
    super().__init__()
    assert target_length > 0
    self._target_length = target_length

  def render(self):
    pending_table = table.Table(
        show_header=False, show_footer=False, show_lines=True, box=box.SQUARE)
    for _ in range(self._target_length):
      pending_table.add_column()
    guess_w_padding = self.pending_guess + " " * (
        self._target_length - len(self.pending_guess))
    pending_table.add_row(*guess_w_padding)
    grid = table.Table.grid()
    grid.add_row(text.Text("Guess:"))
    grid.add_row(pending_table)
    return align.Align(grid, align="center")

  def add_letter(self, letter: str):
    if len(self.pending_guess) < self._target_length:
      letter = util.clean_text(letter)
      check.validate_char(letter)
      self.pending_guess += letter

  def remove_letter(self):
    if len(self.pending_guess) > 0:
      self.pending_guess = self.pending_guess[:-1]

  def clear_guess(self):
    self.pending_guess = ""


class TitlePanel(widget.Widget):
  message = reactive.Reactive("Tordle!")
  style = "bold"

  def render(self):
    return text.Text(self.message, style=self.style)

  def set_victory(self):
    self.style = "green bold"
    self.message = "Victory!"

  def set_defeat(self, target: str):
    self.style = "red bold"
    self.message = f"It was '{target}'"


class ErrorPanel(widget.Widget):
  message = reactive.Reactive("")

  def render(self):
    return text.Text(self.message)


class RootGrid(widget.Widget):

  def __init__(self, sess: session.Session):
    super().__init__()
    self._session = sess
    self.guess_table = GuessTable(sess)
    self.pending_guess = PendingGuessPanel(len(sess.target))
    self.error_panel = ErrorPanel()
    self.title_panel = TitlePanel()
    self._components = [
        self.title_panel,
        self.error_panel,
        self.guess_table,
        self.pending_guess,
    ]

  def render(self):
    grid = table.Table.grid(expand=True)
    grid.add_column(justify="center")
    for c in self._components:
      grid.add_row(c)
    return grid



def clean_text(text: str) -> str:
  return text.strip().upper()


_DEBUG = False


def enable_debug():
  global _DEBUG
  _DEBUG = True


def debugging() -> bool:
  return _DEBUG


def dprint(*args, **kwargs):
  if _DEBUG:
    print("LOG:", *args, **kwargs)



import bisect
import pathlib
import random
from collections import defaultdict
from typing import Iterable, Optional

from . import util

_DEFAULT_WORD_LIST = pathlib.Path(__file__).parent.joinpath("usa_english.txt")


class WordList:

  def __init__(self, words: Optional[Iterable[str]] = None):
    if words is None:
      words = _load_default_word_list()
    self._count = 0
    self._words_by_size = defaultdict(list)
    for word in words:
      self._words_by_size[len(word)].append(util.clean_text(word))
      self._count += 1
    # Sort each list so we can binary search for words later.
    for size in self._words_by_size:
      self._words_by_size[size].sort()

  def __len__(self):
    return self._count

  def get_random_word(self, desired_length: int) -> Optional[str]:
    """Returns a word from the input list of the desired length.

        Returns none if no words of the desired length are available."""
    if desired_length not in self._words_by_size:
      return None
    return random.choice(self._words_by_size[desired_length])

  def __contains__(self, query: str) -> bool:
    query = util.clean_text(query)
    words = self._words_by_size[len(query)]
    if not words:
      return False
    idx = bisect.bisect_left(words, query)
    if idx < 0 or idx >= len(words):
      return False
    return words[idx] == query


def _load_default_word_list():
  assert _DEFAULT_WORD_LIST.is_file()
  with open(_DEFAULT_WORD_LIST) as word_file:
    return [util.clean_text(l) for l in word_file]



import rich
from rich import layout, live, prompt, table

from . import check, session, util


class SessionRenderer():

  def __init__(self, sess: session.Session):
    self._session = sess
    self._live = None

  def _generate_guess_table(self):
    guess_table = table.Table()
    target_length = len(self._session.target)
    for _ in range(target_length):
      guess_table.add_column(justify="center", width=1)
    total_guesses = self._session.total_guesses
    for idx in range(total_guesses):
      if idx < len(self._session.guess_history):
        guess_table.add_row(*list(self._session.guess_history[idx]))
      else:
        empty_guess = [" "] * target_length
        guess_table.add_row(*empty_guess)
    return guess_table

  def __enter__(self):
    self._live = live.Live(
        self._generate_guess_table(), transient=True, screen=True)
    self._live.__enter__()
    return self

  def __exit__(self, *args, **kwargs):
    self._live.__exit__(*args, **kwargs)
    self._live = None

  def _validate_context(self):
    if self._live is None:
      raise ValueError("SessionRenderer should be invoked with 'with'")

  @property
  def console(self):
    self._validate_context()
    return self._live.console

  def print(self, msg: str):
    self.console.print(msg)

  def prompt_guess(self) -> str:
    self._validate_context()
    while True:
      try:
        guess = prompt.Prompt.ask(">", console=self.console)
        check.validate_guess(guess, len(self._session.target))
        return guess
      except ValueError as e:
        self.print(e)

  def update(self):
    self._live.update(self._generate_guess_table())
    # if self._session.status == session.SessionStatus.ACTIVE:
    # rich.print(self._generate_guess_table())
    # elif self._session.status == session.SessionStatus.DEFEAT:
    # print("DEFEAT")
    # elif self._session.status == session.SessionStatus.VICTORY:
    # print("VICTORY")
    # elif self._session.status == session.SessionStatus.ERROR:
    # print("ERROR")
    # else:
    # raise ValueError("Session is in unsupported state: "
    # f"{self._session.status}")


def _hint_to_char(hint: check.HintCategory):
  if hint == check.HintCategory.MISS:
    return "X"
  if hint == check.HintCategory.CLOSE:
    return "!"
  if hint == check.HintCategory.HIT:
    return "$"
  raise ValueError(f"Unsupported hint: {hint}")



import click
from rich import table
from textual import app

from . import render, session, util, widgets, word_list


class TordleApp(app.App):

  def __init__(self, *args, target_length: int, total_guesses: int, **kwargs):
    super().__init__(*args, **kwargs)
    assert target_length >= 0
    assert total_guesses >= 0
    self._total_guesses = total_guesses
    self._target_length = target_length
    self._words = word_list.WordList()
    self._session = session.Session(
        target=self._words.get_random_word(self._target_length),
        total_guesses=self._total_guesses,
        words=self._words)
    self._root_grid = widgets.RootGrid(self._session)

  async def on_mount(self):
    await self.view.dock(self._root_grid)

  def on_key(self, event):
    if self._session.status != session.SessionStatus.ACTIVE:
      return
    try:
      if event.key == "enter":
        self._session.guess(self._root_grid.pending_guess.pending_guess)
        self._root_grid.pending_guess.clear_guess()
      if event.key in {"ctrl+h", "delete"}:
        self._root_grid.pending_guess.remove_letter()
      elif len(event.key) == 1:
        self._root_grid.pending_guess.add_letter(event.key)
      if self._session.status == session.SessionStatus.VICTORY:
        self._root_grid.title_panel.set_victory()
      elif self._session.status == session.SessionStatus.DEFEAT:
        self._root_grid.title_panel.set_defeat(self._session.target)
      self._root_grid.error_panel.message = ""
    except ValueError as e:
      self._root_grid.error_panel.message = str(e)
    finally:
      self._root_grid.refresh()


@click.command()
@click.option(
    "-l",
    "--target-length",
    default=5,
    help="The number of letters in the target word.")
@click.option(
    "-g",
    "--total-guesses",
    default=6,
    help="The number of guesses you have to get the right word.")
@click.option(
    "--debug/--no-debug",
    default=False,
    help="If enabled, we print everying, incluing the answer.")
def main(debug, **kwargs):
  if debug:
    util.enable_debug()
  TordleApp.run(**kwargs)


if __name__ == "__main__":
  main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tordle']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'isort>=5.10.1,<6.0.0',
 'rich>=11.0.0,<12.0.0',
 'textual>=0.1.14,<0.2.0',
 'toml>=0.10.2,<0.11.0',
 'yapf>=0.32.0,<0.33.0']

entry_points = \
{'console_scripts': ['tordle = tordle.tordle:main']}

setup_kwargs = {
    'name': 'tordle',
    'version': '0.1.0',
    'description': 'A command-line word game.',
    'long_description': None,
    'author': 'Justin Sybrandt',
    'author_email': 'justin@sybrandt.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
